/****************************************************************************/
/*                       24HOP 2015 - Summit Preview                        */
/*                                                                          */
/*                  Written by Dmitri V. Korotkevitch                       */
/*                      http://aboutsqlserver.com                           */
/*                        dk@aboutsqlserver.com                             */
/****************************************************************************/
/*                         Parameters Sniffing                              */
/****************************************************************************/

use SqlServerInternals
go

set nocount on
go

if exists(
	select * 
	from sys.tables t join sys.schemas s on t.schema_id = s.schema_id
	where s.name = 'dbo' and t.name = 'People'
)
	drop table dbo.People
go

create table dbo.People
(
	ID int not null identity(1,1),
	Name varchar(100) not null,
	Age tinyint not null,
	Country varchar(100) not null,

	constraint PK_People
	primary key clustered(ID)
)    
go

;WITH N1(C) AS (SELECT 0 UNION ALL SELECT 0) -- 2 rows
,N2(C) AS (SELECT 0 FROM N1 AS T1 CROSS JOIN N1 AS T2) -- 4 rows
,N3(C) AS (SELECT 0 FROM N2 AS T1 CROSS JOIN N2 AS T2) -- 16 rows
,N4(C) AS (SELECT 0 FROM N3 AS T1 CROSS JOIN N3 AS T2) -- 256 rows
,N5(C) AS (SELECT 0 FROM N4 AS T1 CROSS JOIN N4 AS T2) -- 65,535 rows
,IDs(ID) AS (SELECT ROW_NUMBER() OVER (ORDER BY (SELECT NULL)) FROM N5)
insert into dbo.People(Name, Age, Country)
	select convert(varchar(100),ID), ID % 90 + 1, 'USA'
	from IDs
go

;WITH N1(C) AS (SELECT 0 UNION ALL SELECT 0) -- 2 rows
,N2(C) AS (SELECT 0 FROM N1 AS T1 CROSS JOIN N1 AS T2) -- 4 rows
,N3(C) AS (SELECT 0 FROM N2 AS T1 CROSS JOIN N2 AS T2) -- 16 rows
,IDs(ID) AS (SELECT ROW_NUMBER() OVER (ORDER BY (SELECT NULL)) FROM N3)
insert into dbo.People(Name, Age, Country)
	select convert(varchar(100),ID), ID % 90 + 1, 'CANADA'
	from IDs
go

create nonclustered index IDX_People_Country
on dbo.People(Country)
go

dbcc show_statistics('dbo.People','IDX_People_Country')
go

if exists(
	select * 
	from sys.procedures p join sys.schemas s on p.schema_id = s.schema_id
	where s.name = 'dbo' and p.name = 'GetPeopleAge'
)
	drop proc dbo.GetPeopleAge
go

create proc dbo.GetPeopleAge(
	@Country varchar(100)
)
as
	select avg(Age)
	from dbo.People
	where Country = @Country
go

set statistics io on
go

-- Enable Actual Execution Plan

dbcc freeproccache
go

exec dbo.GetPeopleAge @Country = 'USA'
go

exec dbo.GetPeopleAge @Country = 'CANADA'
go

-- Parameter sniffing
dbcc freeproccache
go

exec dbo.GetPeopleAge @Country = 'CANADA'
go

exec dbo.GetPeopleAge @Country = 'USA'
go




-- Solution 1 - Recompilation
alter proc dbo.GetPeopleAge(
	@Country varchar(100)
)
as
	select avg(Age)
	from dbo.People
	where Country = @Country
	option (recompile)
go

dbcc freeproccache
go

exec dbo.GetPeopleAge @Country = 'CANADA'
go

exec dbo.GetPeopleAge @Country = 'USA'
go


-- Solution 2 - Optimize For parameter value
alter proc dbo.GetPeopleAge(
	@Country varchar(100)
)
as
	select avg(Age)
	from dbo.People
	where Country = @Country
	option (optimize for (@Country='USA'))
go

dbcc freeproccache
go

exec dbo.GetPeopleAge @Country = 'CANADA'
go

exec dbo.GetPeopleAge @Country = 'USA'
go


update dbo.People set Country = 'GERMANY' where Country = 'USA'
go


dbcc freeproccache
go

exec dbo.GetPeopleAge @Country = 'USA'
go

exec dbo.GetPeopleAge @Country = 'CANADA'
go

exec dbo.GetPeopleAge @Country = 'GERMANY'
go


-- Solution 3 - Optimize For UNKNOWN
alter proc dbo.GetPeopleAge(
	@Country varchar(100)
)
as
	select avg(Age)
	from dbo.People
	where Country = @Country
	option (optimize for (@Country unknown))
go


dbcc freeproccache
go

exec dbo.GetPeopleAge @Country = 'USA'
go

exec dbo.GetPeopleAge @Country = 'CANADA'
go

exec dbo.GetPeopleAge @Country = 'GERMANY'
go

update dbo.People set Country = 'USA' where Country = 'GERMANY'
go

dbcc freeproccache
go

exec dbo.GetPeopleAge @Country = 'USA'
go

exec dbo.GetPeopleAge @Country = 'CANADA'
go

exec dbo.GetPeopleAge @Country = 'GERMANY'
go


-- Solution 4 - Local variables (same as OPTIMIZE for UNKNOWN)
alter proc dbo.GetPeopleAge(
	@Country varchar(100)
)
as
	declare
		@CountryLocal varchar(100)

	set @CountryLocal = @Country;

	select avg(Age)
	from dbo.People
	where Country = @CountryLocal
go

dbcc freeproccache
go

exec dbo.GetPeopleAge @Country = 'CANADA'
go

exec dbo.GetPeopleAge @Country = 'USA'
go

exec dbo.GetPeopleAge @Country = 'GERMANY'
go

